{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Author:       Franois PIETTE
Description:  Example of a server object which returns complete file
              as blob field.
              Part of the SrvTst sample application server.
Creation:     April 24, 1998
Version:      1.00
EMail:        http://www.overbyte.be        http://www.rtfm.be/fpiette
              francois.piette@overbyte.be   francois.piette@rtfm.be
                                            francois.piette@pophost.eunet.be
Support:      Use the mailing list midware@elists.org
              Follow "support" link at http://www.overbyte.be for subscription.
Legal issues: Copyright (C) 1998-2002 by Franois PIETTE
              Rue de Grady 24, 4053 Embourg, Belgium. Fax: +32-4-365.74.56
              <francois.piette@overbyte.be>

              This software is provided 'as-is', without any express or
              implied warranty. In no event will the author be held liable
              for any damages arising from the use of this software.

              Permission is granted to anyone to use this software and or any
              derived or altered versions for any purpose, excluding commercial
              applications. You can use this software for personal use only.
              You may distribute it freely untouched.
              The following restrictions applies:

              1. The origin of this software must not be misrepresented, you
                 must not claim that you wrote the original software.

              2. If you use this software in a product, an acknowledgment in
                 the product documentation and displayed on screen is required.
                 The text must be: "This product is based on MidWare. Freeware
                 source code is available at http://www.overbyte.be."

              3. Altered source versions must be plainly marked as such, and
                 must not be misrepresented as being the original software.

              4. This notice may not be removed or altered from any source
                 distribution and must be added to the product documentation.

Updates:

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
unit SrvObj5;

interface

uses
    SysUtils, Classes, RBroker, RFormat;

type
    TServerObjectGETFILE = class(TServerObject)
    public
        procedure   Execute; override;
    end;

implementation

{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TServerObjectGETFILE.Execute;
var
    Data     : TFileStream;
    FileName : String;
begin
    { The RequestBuffer is positionned on first record on startup }
    FileName := Trim(FRequestBuffer.Fields[0]);

    if FileName = '' then begin
        FResponseBuffer.WriteFields(TRUE, ['Missing file name']);
        FResultStatus := 400;
        Finish;
    end;

    try
        Data := TFileStream.Create(FileName, fmOpenRead + fmShareDenyNone);
        Data.Seek(0, 0);
        FResponseBuffer.WriteFields(TRUE, [FileName]);
        FResponseBuffer.WriteStreamField(FALSE, mwString, Data);
        Data.Destroy;
        FResultStatus := 200;
    except
        on E:Exception do begin
            FResponseBuffer.WriteFields(TRUE, ['Error',
                                               E.Message]);
            FResultStatus := 400;
        end;
    end;
    Finish;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

end.


